import struct
import numpy as np
from NEDAS.grid.grid_regular import RegularGrid
from NEDAS.models.topaz.abfile import ABFileBathy
from NEDAS.models.topaz.confmap import ConformalMapping

def get_topaz_grid(grid_info_file) -> RegularGrid:
    proj = ConformalMapping.init_from_file(grid_info_file)

    ##the coordinates in topaz model native grid
    ii, jj = np.meshgrid(np.arange(proj._ires), np.arange(proj._jres))
    x = ii * proj._dx
    y = jj * proj._dy

    return RegularGrid(proj, x, y, distance_type='spherical')

def get_depth(depthfile, grid):
    f = ABFileBathy(depthfile, 'r', idm=grid.nx, jdm=grid.ny)
    depth = f.read_field('depth')
    f.close()

    d = -depth.data  ##depth is negative for ocean variables
    
    mask = depth.mask
    d[mask] = np.nan
    
    return d, mask

def get_mean_ssh(meansshfile, grid):
    with open(meansshfile, 'rb') as f:
        f.seek(4)
        data = f.read(8 * grid.nx * grid.ny)
        data = struct.unpack('>'+'d'*grid.nx*grid.ny, data)
    
        meanssh = np.array(data).reshape(grid.ny, grid.nx)
    
    return meanssh

##some variables will need (de)staggering in topaz grid:
##---                *--*--*
##---                |  |  |
##--- stencil:       u--p--*
##---                |  |  |
##---                q--v--*
##these two functions are uniq to topaz

def stagger(dat, vtype):
    ##stagger u,v for C-grid configuration
    dat_stag = dat.copy()
    if vtype == 'u':
        dat_stag[:, 1:] = 0.5*(dat[:, :-1] + dat[:, 1:])
        dat_stag[:, 0] = 3*dat[:, 1] - 3*dat[:, 2] + dat[:, 3]
    elif vtype == 'v':
        dat_stag[1:, :] = 0.5*(dat[:-1, :] + dat[1:, :])
        dat_stag[0, :] = 3*dat[1, :] - 3*dat[2, :] + dat[3, :]
    return dat_stag

def destagger(dat_stag, vtype):
    ##destagger u,v from C-grid
    dat = dat_stag.copy()
    if vtype == 'u':
        dat[:, :-1] = 0.5*(dat_stag[:, :-1] + dat_stag[:, 1:])
        dat[:, -1] = 3*dat_stag[:, -2] - 3*dat_stag[:, -3] + dat_stag[:, -4]
    elif vtype == 'v':
        dat[:-1, :] = 0.5*(dat_stag[:-1, :] + dat_stag[1:, :])
        dat[-1, :] = 3*dat_stag[-2, :] - 3*dat_stag[-3, :] + dat_stag[-4, :]
    return dat




import os
import yaml
import cftime
import numpy as np
from pyproj import Proj
from NEDAS.grid import Grid
from NEDAS.utils.netcdf_lib import nc_write_var
from NEDAS.utils.conversion import ensure_list, proj2dict, s2t, dt1h
from NEDAS.utils.shell_utils import makedir

def get_task_list(c, **kwargs):
    """get a list of tasks to be done, as unique kwargs to be passed to run()"""
    # parse kwargs
    # load the default config first
    default_config_file = os.path.join(os.path.dirname(__file__), 'default.yml')
    with open(default_config_file, 'r') as f:
        kwargs_from_default = yaml.safe_load(f)
    kwargs_from_config_file = {}
    if 'config_file' in kwargs and kwargs['config_file'] is not None:
        with open(kwargs['config_file'], 'r') as f:
            kwargs_from_config_file = yaml.safe_load(f)
    # overwrite with kwargs
    kwargs = {**kwargs_from_default, **kwargs_from_config_file, **kwargs}

    ##generate task list
    tasks = []
    for member in range(c.nens):
        for vname in ensure_list(kwargs['variables']):
            tasks.append({**kwargs, 'variable': vname, 'member': member})
    return tasks

def get_file_list(c, **kwargs):
    """Return a list of files to be opened for collective i/o"""
    if 'time' in kwargs:
        time_start = s2t(kwargs['time'])
    else:
        time_start = c.time

    files = []
    for member in range(c.nens):
        file = kwargs['file'].format(time=time_start, member=member+1)
        if file not in files:
            files.append(file)
    return files

def run(c, **kwargs):
    """
    Run diagnostics: convert model restart variables into netcdf files, given formatting options in kwargs
    """
    vname = kwargs['variable']
    member = kwargs['member']
    model_name = kwargs['model_src']
    model = c.models[model_name]

    grid_def = kwargs['grid_def']
    if grid_def:
        proj = Proj(grid_def['proj'])
        xmin = grid_def['xmin']
        xmax = grid_def['xmax']
        ymin = grid_def['ymin']
        ymax = grid_def['ymax']
        dx = grid_def['dx']
        centered = grid_def.get('centered', False)
        grid = Grid.regular_grid(proj, xmin, xmax, ymin, ymax, dx, centered)
    else:
        grid = model.grid
    proj_params = proj2dict(grid.proj)
    x = grid.x[0, :] / 1e5
    y = grid.y[:, 0] / 1e5  ##convert to 100km units
    lon, lat = grid.proj(grid.x, grid.y, inverse=True)

    if 'time' in kwargs:
        time_start = s2t(kwargs['time'])
    else:
        time_start = c.time
    dt_hours = kwargs.get("dt_hours", model.output_dt)
    forecast_hours = kwargs.get("forecast_hours", c.cycle_period)
    time_units = kwargs.get('time_units', 'seconds since 1970-01-01T00:00:00+00:00')
    time_calendar = kwargs.get('time_calendar', 'standard')
    t_steps = range(0, forecast_hours, dt_hours)
    path = c.forecast_dir(time_start, model_name)

    for n_step, t_step in enumerate(t_steps):
        t = time_start + t_step * dt1h

        if c.debug:
            print(f"PID {c.pid:4} convert_output on variable '{vname:20}' for {model_name:10} member {member+1:3} at {t}", flush=True)
        # read the variable from the model restart file
        rec = model.variables[vname]
        file = kwargs['file'].format(time=time_start, member=member+1)
        makedir(os.path.dirname(file))

        lon_name: str = kwargs.get('lon_name', 'lon')
        lat_name: str = kwargs.get('lat_name', 'lat')
        x_name: str = kwargs.get('x_name', 'x')
        y_name: str = kwargs.get('y_name', 'y')
        time_name: str = kwargs.get('time_name', 'time')
        recno = {}
        recno[time_name] = n_step
        levels = rec['levels']
        is_vector = rec['is_vector']
        for k in levels:
            # read the field from model restart file
            model.read_grid(path=path, name=vname, time=t, member=member, k=k)
            fld = model.read_var(path=path, name=vname, time=t, member=member, k=k)
            model.grid.set_destination_grid(grid)

            # convert to output grid
            fld_ = model.grid.convert(fld, is_vector=is_vector)
            # build dimension records
            dims = {}
            dims[time_name] = None  ##make time dimension unlimited in nc file
            k_name = kwargs.get('k_name')
            if len(levels) > 1:
                dims[k_name] = None  ##add level dimension (unlimited) if there are multiple levels
            dims[y_name] = grid.ny
            dims[x_name] = grid.nx
            # output the variable
            if len(levels) > 1:
                recno[k_name] = list(levels).index(k)
            # variable attr
            attr = {'standard_name':vname,
                    'units':rec['units'],
                    'grid_mapping': proj_params['projection'],
                    'coordinates': f"{lon_name} {lat_name}",
                    }
            if is_vector:
                for i in range(2):
                    if isinstance(rec['name'], tuple):
                        rec_name = rec['name'][i]
                    else:
                        rec_name = rec['name']+'_'+(x_name, y_name)[i]
                    nc_write_var(file, dims, rec_name, fld_[i,...], recno=recno, attr=attr, comm=c.comm)
            else:
                nc_write_var(file, dims, rec['name'], fld_, recno=recno, attr=attr, comm=c.comm)

        # output the dimension variables
        time = cftime.date2num(t, units=time_units)
        time_attr = {'long_name': 'forecast time', 'units': time_units, 'calendar': time_calendar}
        nc_write_var(file, {time_name:None}, time_name, time, dtype=float, recno=recno, attr=time_attr, comm=c.comm)
        nc_write_var(file, {x_name:grid.nx}, x_name, x, attr={'standard_name':'projection_x_coordinate', 'units':'100 km'}, comm=c.comm)
        nc_write_var(file, {y_name:grid.ny}, y_name, y, attr={'standard_name':'projection_y_coordinate', 'units':'100 km'}, comm=c.comm)
        nc_write_var(file, {y_name:grid.ny, x_name:grid.nx}, lon_name, lon, attr={'standard_name':'longitude', 'units':'degrees_east'}, comm=c.comm)
        nc_write_var(file, {y_name:grid.ny, x_name:grid.nx}, lat_name, lat, attr={'standard_name':'latitude', 'units':'degrees_north'}, comm=c.comm)
        # output projection info
        nc_write_var(file, {}, proj_params['projection'], np.array([1]), attr=proj_params, comm=c.comm)



import sys
import subprocess

class JobSubmitter:
    def __init__(self, **kwargs):
        """
        Input kwargs:
        -job_name: str, name of the job
        -nproc: int, number of processors requested
        -offset: int, number of processors to offset from the start
        -ppn: int, number of processors per compute node
        -host: str, host machine name
        -project: str or None, project id for allocation
        -queue: str or None, which job queue to use
        -walltime: int, maximum run time allowed (in seconds)
        -check_dt: int, time interval (in seconds) to check job status
        -run_separate_jobs: bool, if true, submit each job separately to the scheduler, otherwise run in the same allocation as job steps.
        -use_job_array: bool, if true, use job array for ensemble members
        -array_size: int, size of the job array
        -parallel_mode
        -debug
        """
        ## setting default parameters if not specified in kwargs
        self.job_name = kwargs.get('job_name', 'run')
        self.run_dir = kwargs.get('run_dir', '.')
        self._nproc = kwargs.get('nproc', 1)
        self._ppn = kwargs.get('ppn', 1)
        self._offset = kwargs.get('offset', 0)
        self.host = kwargs.get('host', 'localhost').lower()
        self.project = kwargs.get('project')
        self.queue = kwargs.get('queue')
        self.walltime = kwargs.get('walltime', 3600)
        self.check_dt = kwargs.get('check_dt', 20)
        self.run_separate_jobs = kwargs.get('run_separate_jobs', False)
        self.use_job_array = kwargs.get('use_job_array', False)
        self.array_size = kwargs.get('array_size', 1)
        self.parallel_mode = kwargs.get('parallel_mode', 'mpi')
        self.debug = kwargs.get('debug', False)

    @property
    def nproc(self) -> int:
        """
        Number of requested processors for the job
        """
        return self._nproc

    @nproc.setter
    def nproc(self, value):
        self._nproc = value

    @property
    def ppn(self) -> int:
        """
        Number of processors per compute node requested for the job
        """
        return self._ppn

    @ppn.setter
    def ppn(self, value):
        self._ppn = value

    @property
    def offset(self) -> int:
        """
        Number of processors to skip from the beginning for the job
        This allows different jobs to spawm the total available nproc in the allocation
        Discarded if run_separate_jobs
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = value

    @property
    def nnode(self) -> int:
        """
        Number of compute nodes for the job
        """
        return (self._nproc + self._ppn - 1) // self._ppn

    @property
    def offset_node(self) -> int:
        """
        Number of compute nodes to skip from the beginning
        """
        return self._offset // self._ppn

    @property
    def nproc_avail(self) -> int:
        """
        Number of available processors on a host machine (only used when not run_separate_jobs)
        Vanila JobSubmitter assumes that requested nproc is always available
        This should be redefined in subclasses to machine specific behavior
        """
        return self.nproc + self.offset

    @property
    def nnode_avail(self) -> int:
        """
        Number of available compute nodes on a host machine (only used when not run_separate_jobs)
        """
        return self.nnode + self.offset_node

    @property
    def ppn_avail(self) -> int:
        """
        Number of available processors per compute node (only used when not run_separate_jobs)
        """
        return self.ppn

    @property
    def execute_command(self) -> str:
        """
        Execute command for running the job on the host machine, replacing 'JOB_EXECUTE' in 'commands'
        Vanila JobSubmitter will just run "mpirun -np nproc ...", and discard the ppn and offset settings
        """
        if self.parallel_mode == 'mpi':
            return f"mpirun -np {self.nproc}"
        elif self.parallel_mode == 'openmp':
            return f"export OMP_NUM_THREADS={self.nproc};"
        else:
            raise ValueError(f"unknown parallel_mode '{self.parallel_mode}'")

    @property
    def job_array_index_name(self) -> str:
        """
        Job array index variable name for the host machine, replacing 'JOB_ARRAY_INDEX' in 'commands'
        """
        return 'JOB_ARRAY_INDEX'

    def parse_commands(self, commands) -> str:
        commands = commands.replace('JOB_EXECUTE', self.execute_command)
        commands = commands.replace('JOB_ARRAY_INDEX', self.job_array_index_name)
        return commands

    def check_resources(self) -> None:
        ##check if requested resource is available
        assert self.nproc+self.offset <= self.nproc_avail, f"Requested nproc={self.nproc} and offset={self.offset} exceeds nproc_avail={self.nproc_avail}"
        assert self.nnode+self.offset_node <= self.nnode_avail, f"Requested nnode={self.nnode} and offset_node={self.offset_node} exceeds nnode_avail={self.nnode_avail}"

    def run_job_as_step(self, commands) -> None:
        """
        Run 'commands' from within a job allocation
        Use nproc processors starting from the offset+1 processor of the allocation
        """
        self.check_resources()

        commands = self.parse_commands(commands)
        if self.debug:
            print("JobSubmitter run command as step: ", commands, flush=True)

        p = subprocess.Popen(commands, shell=True, text=True, bufsize=1)
        p.wait()

        ##handle error
        if p.returncode != 0:
            print(f"JobSubmitter: job '{self.job_name}' exited with error")
            sys.exit(1)

    def submit_job_and_monitor(self, commands) -> None:
        raise NotImplementedError("run_separate_jobs=True is not availalbe without a scheduler")

    def run(self, commands) -> None:
        """
        Top level run method
        Input commands: str, shell commands to be run by the job submitter
        In the commands string, 'JOB_EXECUTE' will be replaced by the correct execute_command,
        and 'JOB_ARRAY_INDEX' will be replaced by the scheduler's index variable name to perform array jobs.
        """
        if self.run_separate_jobs:
            self.submit_job_and_monitor(commands)

        else:
            self.run_job_as_step(commands)




import numpy as np
import glob, os
from pyproj import Proj
from scipy.interpolate import LinearNDInterpolator
from netCDF4 import Dataset
from datetime import datetime, timedelta
from NEDAS.grid import Grid
from .proj import lonlat2xy, xy2lonlat
from NEDAS.models import Model

class NorESM(Model):
    pass

levels = np.arange(1, 31, 1)
variables = {'ocean_velocity': {'name':('u', 'v'), 'dtype':'float', 'is_vector':True, 'restart_dt':1440, 'levels':levels, 'units':'m/s'},
             'ocean_temp': {'name':'temp', 'dtype':'float', 'is_vector':False, 'restart_dt':1440, 'levels':levels, 'units':'m/s'},
             'ocean_saln': {'name':'saln', 'dtype':'float', 'is_vector':False, 'restart_dt':1440, 'levels':levels, 'units':'m/s'},
            }

##staggering types for native model variables: 'p','u','v','q'
stagger = {'u':'u', 'v':'v', 'temp':'p', 'saln':'p'}

def filename(path, **kwargs):
    """
    Parse kwargs and find matching filename
    for keys in kwargs that are not set, here we define the default values
    key values in kwargs will also be checked for erroneous values
    """
    if 'time' in kwargs and kwargs['time'] is not None:
        assert isinstance(kwargs['time'], datetime), 'time shall be a datetime object'
        tstr = kwargs['time'].strftime('%Y%m%d%H%M')
    else:
        tstr = '????????'
    if 'member' in kwargs and kwargs['member'] is not None:
        assert kwargs['member'] >= 0, 'member index shall be >= 0'
        mstr = '_mem{:03d}'.format(kwargs['member']+1)
    else:
        mstr = ''

    ##get a list of filenames with matching kwargs
    search = path+'/'+'output'+tstr+mstr+'.nc'
    flist = glob.glob(search)
    assert len(flist)>0, 'no matching files found: '+search
    ##typically there will be only one matching file given the kwargs,
    ##if there is a list of matching files, we return the first one
    return flist[0]

def grid_info(grid_file, grid_type, scale_x=1, scale_y=1, stagger='p'):
    """
    Fetch grid info from a given grid_file, or generate from locally stored data.

    Args:
        grid_file (str): The path to the grid.nc file containing plat,plon...
        grid_type (str): Type of grid, 'bipolar' or 'tripolar'.
        scale_x (float): Resolution scaling in x direction.
        scale_y (float): Resolution scaling in y direction.
        stagger (str): Staggering type, 'p', 'u', 'v', or 'q'.

    Returns:
        tuple: A tuple containing:
            - lon (np.ndarray): Longitude defined on the grid points, of shape (ny, nx).
            - lat (np.ndarray): Latitude defined on the grid points
            - x (np.ndarray): X-coordinates of the grid points.
            - y (np.ndarray): Y-coordinates of the grid points.
            - neighbor (np.ndarray): Neighbor indices of shape (2, 4, ny, nx).
                For each point (:code:`j`, :code:`i`) in (ny,nx), :code:`grid_neighbors[0,:,j,i]` are the :code:`j`-indices for the
                4 neighbors (east, north, west and south) to point (:code:`j`, :code:`i`) and :code:`grid_neighbors[1,:,j,i]` are the corresponding :code:`i`-indices
    """
    ##if provided grid_file doesn't exist, try use locally stored file
    if os.path.exists(grid_file):
        filename = grid_file
    else:
        if grid_type=='bipolar':
            filename = __path__[0]+'/grid_bipolar.nc'
        elif grid_type=='tripolar':
            filename = __path__[0]+'/grid_tripolar.nc'
        else:
            raise ValueError('cannot find grid file for grid_type '+grid_type)
    # print('reading grid file: '+filename)

    with Dataset(filename) as f:
        if stagger in ['p', 'u', 'v', 'q']:
            lon = f[stagger+'lon'][...].data
            lat = f[stagger+'lat'][...].data
        else:
            raise ValueError('unknown staggering type '+stagger)

    if os.path.exists(grid_file):
        ##just get lon, lat from provided grid_file, strip padded rows
        if grid_type=='bipolar':
            grid_lon, grid_lat = lon, lat

        elif grid_type=='tripolar':
            grid_lon, grid_lat = lon[:-1, :], lat[:-1, :]

    else:
        ##refine/coarsen the lon,lat grid to the desired scaling
        ##note: this part is only for testing, usually a grid_file is provided directly
        ##      the manual refining/coarsening here is not so accurate

        ##helper proj to avoid interpolating coordinates (lon,lat) that are not continuous
        ##the north pole stereographic projection is good, since south pole
        ##is not part of the grid anyway, on hproj the coordinates are continuous
        hproj = Proj("+proj=stere +lon_0=0 +lat_0=90")

        if grid_type=='bipolar':
            ##pad orig grid to prepare for interpolation
            ny, nx = lon.shape
            lon_, lat_ = np.zeros((ny, nx+1)), np.zeros((ny, nx+1))
            lat_[:, :-1] = lat
            lat_[:, -1] = lat[:, 0]  ##cyclic east-west boundary
            lon_[:, :-1] = lon
            lon_[:, -1] = lon[:, 0]  ##cyclic east-west boundary

            hx, hy = hproj(lon_, lat_)
            xi, yi = np.meshgrid(np.arange(nx+1), np.arange(ny))
            xy2hx = LinearNDInterpolator(list(zip(xi.flatten(), yi.flatten())), hx.flatten())
            xy2hy = LinearNDInterpolator(list(zip(xi.flatten(), yi.flatten())), hy.flatten())

            x_ = np.arange(0, nx, scale_x)
            y_ = np.arange(0, ny-1+scale_y, scale_y)
            xo, yo = np.meshgrid(x_, y_)
            grid_lon, grid_lat = hproj(xy2hx(xo, yo), xy2hy(xo, yo), inverse=True)

        elif grid_type=='tripolar':
            ##pad orig grid to prepare for interpolation
            ny, nx = lon.shape
            lon_, lat_ = np.zeros((ny+1, nx+2)), np.zeros((ny+1, nx+2))
            lat_[1:, 1:-1] = lat
            lat_[1:, 0] = lat[:, -1]  ##cyclic west boundary
            lat_[1:, -1] = lat[:, 0]  ##cyclic east boundary
            lat_[0, :] = -80.2851  ##south boundary, add another latitude band
                                   ##better: extrapolated from lat[1:5,:]
            lon_[1:, 1:-1] = lon
            lon_[1:, 0] = lon[:, -1]  ##cyclic west boundary
            lon_[1:, -1] = lon[:, 0]  ##cyclic east boundary
            lon_[0, :] = lon_[1, :]   ##south boundary, add another latitude band

            hx, hy = hproj(lon_, lat_)
            xi, yi = np.meshgrid(np.arange(nx+2), np.arange(ny+1))
            xy2hx = LinearNDInterpolator(list(zip(xi.flatten(), yi.flatten())), hx.flatten())
            xy2hy = LinearNDInterpolator(list(zip(xi.flatten(), yi.flatten())), hy.flatten())

            x_ = np.arange(1+(scale_x-1)*0.5, nx+1+(scale_x-1)*0.5, scale_x)
            yc = 148   ##critical latitude 49 degree index in 0-192
            yp1 = np.arange(scale_y, yc, scale_y)
            yp2st = yp1[-1] + scale_y
            yp2ed = ny - 0.5
            yp2 = np.arange(yp2st, yp2ed, (yp2ed - scale_y/2 - yp2st)/((ny-1)/scale_y - yp1.size-1))
            y_ = np.hstack([yp1, yp2])
            xo, yo = np.meshgrid(x_, y_)
            grid_lon, grid_lat = hproj(xy2hx(xo, yo), xy2hy(xo, yo), inverse=True)

    ny, nx = grid_lon.shape
    x, y = np.meshgrid(np.arange(nx), np.arange(ny))
    grid_x, grid_y = x, y

    ##neighbor indices
    grid_neighbors = np.zeros((2, 4, ny, nx), dtype='int')
    if grid_type=='bipolar':
        ##y component
        grid_neighbors[0, 0, ...] = y                       ##east
        grid_neighbors[0, 1, :-1, :] = y[1:, :]             ##north
        grid_neighbors[0, 1, -1, :] = y[-1, :]
        grid_neighbors[0, 2, ...] = y                       ##west
        grid_neighbors[0, 3, 0, :] = y[0, :]                ##south
        grid_neighbors[0, 3, 1:, :] = y[:-1, :]
        ##x component
        grid_neighbors[1, 0, ...] = np.roll(x, -1, axis=1)  ##east
        grid_neighbors[1, 1, :, :] = x                      ##north
        grid_neighbors[1, 2, ...] = np.roll(x, 1, axis=1)   ##west
        grid_neighbors[1, 3, ...] = x                       ##south

    elif grid_type=='tripolar':
        ##y component
        grid_neighbors[0, 0, ...] = y                       ##east
        grid_neighbors[0, 1, :-1, :] = y[1:, :]             ##north
        grid_neighbors[0, 1, -1, :] = y[-2, :]
        grid_neighbors[0, 2, ...] = y                       ##west
        grid_neighbors[0, 3, 0, :] = y[0, :]                ##south
        grid_neighbors[0, 3, 1:, :] = y[:-1, :]
        ##x component
        grid_neighbors[1, 0, ...] = np.roll(x, -1, axis=1)  ##east
        grid_neighbors[1, 1, :-1, :] = x[:-1, :]            ##north
        grid_neighbors[1, 1, -1, :] = x[-1, ::-1]
        grid_neighbors[1, 2, ...] = np.roll(x, 1, axis=1)   ##west
        grid_neighbors[1, 3, ...] = x                       ##south

    return grid_lon, grid_lat, grid_x, grid_y, grid_neighbors

def read_grid(path, **kwargs):
    """ Generate a Grid object for the NorESM model grid """

    ##NorESM uses bipolar and tripolar ocean model grids
    ##See https://expearth.uib.no/?page_id=28 for some explanation
    ##See documentation: https://noresm-docs.readthedocs.io

    if 'grid_type' in kwargs:
        grid_type = kwargs['grid_type']
    else:
        grid_type = 'tripolar'  ##type of grid should match the grid.nc file

    ##scaling of the model grid, 1 correspond to the given grid_file
    ##(locally stored grids: 192x180 tripolar grid, 192x160 bipolar grid)
    ##0.5 means the resolution doubles, 2 means the resolution reduces by half
    if 'scale_x' in kwargs:
        scale_x = kwargs['scale_x']
    else:
        scale_x = 0.5  ##default
    if 'scale_y' in kwargs:
        scale_y = kwargs['scale_y']
    else:
        scale_y = 0.5  ##default

    if not 'stagger' in kwargs:
        kwargs['stagger'] = 'p'

    lon, lat, x, y, neighbors = grid_info(path+'/grid.nc', grid_type, scale_x, scale_y, kwargs['stagger'])

    ##proj function mimics a pyproj.Proj object to convert lon,lat to model x,y
    def proj(xi, yi, inverse=False):
        shape = np.atleast_1d(xi).shape
        xi, yi = np.atleast_1d(xi).flatten(), np.atleast_1d(yi).flatten()
        xo, yo = np.full(xi.size, np.nan), np.full(xi.size, np.nan)
        if not inverse:
            xo, yo = lonlat2xy(lon, lat, x, y, neighbors, xi, yi)
        else:
            xo, yo = xy2lonlat(lon, lat, x, y, neighbors, xi, yi)
        if xo.size == 1:
            return xo.item(), yo.item()
        return xo.reshape(shape), yo.reshape(shape)

    ##set some attributes for proj info
    proj.name = grid_type
    proj.definition = 'proj='+grid_type+' lon_0=70 lat_0=0'

    return Grid(proj, x, y, neighbors=neighbors)

def write_grid(path, **kwargs):
    pass

def read_var(path, grid, **kwargs):
    pass

def write_var():
    pass




import os
import inspect
from abc import ABC, abstractmethod
import numpy as np
from datetime import datetime
from NEDAS.config import parse_config

class Model(ABC):
    """
    Class for configuring and running a model
    """
    def __init__(self, config_file=None, parse_args=False, **kwargs):

        ##parse config file and obtain a list of attributes
        ##get a list of values from default.yml and update with kwargs, save to config_dict
        code_dir = os.path.dirname(inspect.getfile(self.__class__))
        config_dict = parse_config(code_dir, config_file, parse_args, **kwargs)
        for key, value in config_dict.items():
            setattr(self, key, value)

        self.grid = None
        self.z = None

        self.variables = {}

        self.run_process = None
        self.run_status = 'pending'

    def parse_kwargs(self, **kwargs):
        ##args that pinpoints a certain model state variable
        if 'path' not in kwargs:
            kwargs['path'] = '.'  ##default path is current dir

        if 'member' not in kwargs:
            kwargs['member'] = None
        if kwargs['member'] is not None:
            assert kwargs['member'] >= 0, f"member index should be >= 0, got {kwargs['member']}"

        if 'name' not in kwargs:
            kwargs['name'] = list(self.variables.keys())[0]  ##if not specified, use first variable listed
        assert kwargs['name'] in self.variables, f"'{kwargs['name']}' is not defined in model variables"

        if 'time' not in kwargs:
            kwargs['time'] = None
        if kwargs['time'] is not None:
            assert isinstance(kwargs['time'], datetime), "kwargs 'time' is expected to be a datetime object'"

        levels = list(self.variables[kwargs['name']]['levels'])
        if 'k' not in kwargs:
            kwargs['k'] = levels[0]  ##set to the first level if not specified
        assert kwargs['k'] in levels, f"level {kwargs['k']} is not available for variable {kwargs['name']}"

        if 'units' not in kwargs:
            kwargs['units'] = self.variables[kwargs['name']]['units']

        for key in ['restart_dir', 'forecast_period', 'comm']:
            if key not in kwargs:
                kwargs[key] = None

        return kwargs

    @abstractmethod
    def read_grid(self, **kwargs) -> None:
        """
        Read the grid information from the model output.

        Args:
            **kwargs: Keyword arguments for reading the grid.
        """
        pass

    @abstractmethod
    def read_var(self, **kwargs) -> np.ndarray:
        """
        Read a variable from the model output.

        Args:
            **kwargs: Keyword arguments for reading the variable.

        Returns:
            np.ndarray: The read variable.
        """
        pass

    @abstractmethod
    def write_var(self, var, **kwargs) -> None:
        """
        Write a variable to the model output.

        Args:
            var (np.ndarray): The variable to write.
            **kwargs: Keyword arguments for writing the variable.
        """
        pass

    @abstractmethod
    def z_coords(self, **kwargs) -> np.ndarray:
        """
        Get the vertical coordinates of the model.

        Args:
            **kwargs: Keyword arguments for getting the vertical coordinates.

        Returns:
            np.ndarray: The vertical coordinates.
        """
        pass

    @abstractmethod
    def preprocess(self, **kwargs) -> None:
        """
        Preprocess the model data.

        Args:
            **kwargs: Keyword arguments for preprocessing.
        """
        pass

    @abstractmethod
    def postprocess(self, **kwargs) -> None:
        """
        Postprocess the model data.

        Args:
            **kwargs: Keyword arguments for postprocessing.
        """
        pass

    @abstractmethod
    def run(self, **kwargs) -> None:
        """
        Run the model.

        Args:
            **kwargs: Keyword arguments for running the model.
        """
        pass



import unittest
import importlib
from NEDAS.config import Config
from NEDAS.assim_tools.assimilators import registry, get_assimilator

class TestAnalysisScheme(unittest.TestCase):
    def test_assimilator_init(self):
        c = Config()
        for assimilator_name in registry.keys():
            c.assimilator_def['type'] = assimilator_name
            module = importlib.import_module('NEDAS.assim_tools.assimilators.'+assimilator_name)
            assimilator = get_assimilator(c)
            self.assertIsInstance(assimilator, getattr(module, registry[assimilator_name]))

    def test_raise_exception_when_not_implemented(self):
        c = Config()
        with self.assertRaises(NotImplementedError):
            c.assimilator_def['type'] = 'foo'
            get_assimilator(c)

    def test_assimilation_algorithm_implemented(self):
        c = Config()
        for assimilator_name in registry.keys():
            c.assimilator_def['type'] = assimilator_name
            assimilator = get_assimilator(c)

            self.assertTrue(hasattr(assimilator, 'assimilation_algorithm'), "Method 'assimilation_algorithm' not found")

            method = getattr(assimilator, 'assimilation_algorithm')
            self.assertTrue(callable(method), "'assimilation_algorithm' is not callable")

if __name__ == '__main__':
    unittest.main()



import os
import glob
import numpy as np
from datetime import datetime, timedelta
from NEDAS.utils.netcdf_lib import nc_read_var, nc_write_var
from NEDAS.utils.conversion import t2s, s2t, dt1h, units_convert
from NEDAS.utils.shell_utils import run_command, makedir, run_job
from NEDAS.utils.progress import watch_files
from NEDAS.grid import Grid
from NEDAS.models import Model
from .gmshlib import read_mshfile, proj
from .bin_io import read_data, write_data
from .drift_utils import get_deformation_nodes
from .namelist import namelist

class NextsimModel(Model):
    """
    Class for configuring and running the nextsim v1 model (lagrangian version)
    """
    def __init__(self, config_file=None, parse_args=False, **kwargs):

        super().__init__(config_file, parse_args, **kwargs)

        ##Note: we only work with restart files, normal nextsim binfile have some variables names that
        ##are different from restart files, e.g. Concentration instead of M_conc
        self.native_variables = {
            'seaice_conc': {'name':'M_conc', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1 },
            'seaice_thick': {'name':'M_thick', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'m' },
            'seaice_velocity': {'name':'M_VT', 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':[0], 'units':'m/s' },
            'seaice_damage': {'name':'M_damage', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1 },
            'seaice_ridge_ratio': {'name':'M_ridge_ratio', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1 },
            'seaice_conc_young': {'name':'M_conc_young', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1 },
            'seaice_thick_young': {'name':'M_h_young', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'m' },
            'seaice_age': {'name':'M_age', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'year' },
            'seaice_conc_myi': {'name':'M_conc_myi', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1 },
            'seaice_thick_myi': {'name':'M_thick_myi', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':1 },
            'snow_thick': {'name':'M_snow_thick', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'m' },
            'snow_thick_young': {'name':'M_hs_young', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'m' },
            }
        self.diag_variables = {
            'seaice_drift': {'name':'drift', 'dtype':'float', 'is_vector':True, 'dt':self.restart_dt, 'levels':[0], 'units':'km/day' },
            'seaice_deform_div': {'name':'deform_e1', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'1/day' },
            'seaice_deform_shear': {'name':'deform_e2', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'1/day' },
            'seaice_deform_vort': {'name':'deform_e3', 'dtype':'float', 'is_vector':False, 'dt':self.restart_dt, 'levels':[0], 'units':'1/day' },
            }
        self.atmos_forcing_variables = {
            'atmos_surf_velocity': {'name':('x_wind_10m', 'y_wind_10m'), 'is_vector':True, 'dt':self.forcing_dt, 'levels':[0], 'units':'m/s'},
            'atmos_surf_temp': {'name':'air_temperature_2m', 'is_vector':False, 'dt':self.forcing_dt, 'levels':[0], 'units':'K'},
            'atmos_surf_dew_temp': {'name':'dew_point_temperature_2m', 'is_vector':False, 'dt':self.forcing_dt, 'levels':[0], 'units':'K'},
            'atmos_surf_press': {'name':'atm_pressure', 'is_vector':False, 'dt':self.forcing_dt, 'levels':[0], 'units':'Pa'},
            'atmos_precip': {'name':'total_precipitation_rate', 'is_vector':False, 'dt':self.forcing_dt, 'levels':[0], 'units':'kg/m2/s'},
            'atmos_snowfall': {'name':'snowfall_rate', 'is_vector':False, 'dt':self.forcing_dt, 'levels':[0], 'units':'kg/m2/s'},
            'atmos_down_shortwave': {'name':'instantaneous_downwelling_shortwave_radiation', 'is_vector':False, 'dt':self.forcing_dt, 'levels':[0], 'units':'W/m2'},
            'atmos_down_longwave': {'name':'instantaneous_downwelling_longwave_radiation', 'is_vector':False, 'dt':self.forcing_dt, 'levels':[0], 'units':'W/m2'},
            }
        self.variables = {**self.native_variables, **self.diag_variables, **self.atmos_forcing_variables}

        self.grid = None
        self.grid_bank = {}

    def filename(self, **kwargs):
        """
        Get the filename with specified variable name, time, member, etc. 
        """
        kwargs = super().parse_kwargs(**kwargs)
        if kwargs['member'] is not None:
            mstr = '{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''

        if kwargs['name'] in {**self.native_variables, **self.diag_variables}:
            if kwargs['time'] is not None:
                tstr = kwargs['time'].strftime('%Y%m%dT%H%M%SZ')
                return os.path.join(kwargs['path'], mstr, self.restart_input_path, 'field_'+tstr+'.bin')

            else:
                tstr = '*'
                search = os.path.join(kwargs['path'], mstr, self.restart_input_path, 'field_'+tstr+'.bin')
                flist = glob.glob(search)
                assert len(flist)>0, 'no matching files found: '+search
                return flist[0]   ##return the first matching file

        elif kwargs['name'] in self.atmos_forcing_variables:
            return os.path.join(kwargs['path'], mstr, "data", self.atmos_forcing_path, "generic_ps_atm_"+kwargs['time'].strftime('%Y%m%d')+".nc")

    def read_grid_from_mshfile(self, mshfile):
        """
        Read mshfile and update the self.grid object
        """
        grid_info = read_mshfile(mshfile)
        x, y = grid_info['nodes_x'], grid_info['nodes_y']
        triangles = np.array([np.array(el.node_indices) for el in grid_info['triangles']])
        self.grid = Grid(proj, x, y, regular=False, triangles=triangles)
        self.edges = np.array([np.array(el.node_indices) for el in grid_info['edges']])

    def read_grid(self, **kwargs):
        """
        Update self.grid object based on input kwargs
        """
        if kwargs is None:
            self.read_grid_from_mshfile(os.path.join(self.nextsim_mesh_dir, self.msh_filename))

        kwargs = super().parse_kwargs(**kwargs)
        if kwargs['name'] in {**self.native_variables, **self.diag_variables}:
            if 'meshfile' in kwargs:
                meshfile = kwargs['meshfile']
            else:
                meshfile = self.filename(**kwargs).replace('field', 'mesh')
            ###only need to read the uniq grid once, store known meshfile in memory bank
            if meshfile not in self.grid_bank:
                ##read the grid from mesh file and add to grid_bank
                x = read_data(meshfile, 'Nodes_x')
                y = read_data(meshfile, 'Nodes_y')
                elements = read_data(meshfile, 'Elements')
                n_elements = int(elements.size/3)
                triangles = elements.reshape((n_elements, 3)) - 1
                self.grid_bank[meshfile] = Grid(proj, x, y, regular=False, triangles=triangles)
                self.grid_bank[meshfile].id = read_data(meshfile, 'id')
            self.grid = self.grid_bank[meshfile]

        elif kwargs['name'] in self.atmos_forcing_variables:
            self.grid = Grid.regular_grid(proj, -2.5e6, 2.498e6, -2e6, 2.5e6, 3e3, centered=True)

        self.grid.mask = np.full(self.grid.x.shape, False)  ##no grid points should be masked

    def write_grid(self, **kwargs):
        """
        write updated mesh back to mesh file

        Note: now we assume that number of mesh elements and their indices doesn't change!
        only updating the mesh node position x,y
        """
        kwargs = super().parse_kwargs(**kwargs)
        meshfile = self.filename(**kwargs).replace('field', 'mesh')

        write_data(meshfile, 'Nodes_x', self.grid.x)
        write_data(meshfile, 'Nodes_y', self.grid.y)

        elements = (self.grid.tri.triangles + 1).flatten()
        write_data(meshfile, 'Elements', elements)

    def get_boundary_nodes(self):
        edges = set()
        for triangle in self.grid.tri.triangles:
            for i in range(3):
                edge = tuple(sorted((triangle[i], triangle[(i+1) % 3])))
                if edge in edges:
                    edges.remove(edge)  # If seen twice, it's an internal edge
                else:
                    edges.add(edge)
        return np.unique(np.array(list(edges)))  ##the boundary node indices

    def get_neighbor_nodes(self, nodes):
        neighbors = set()
        for i in nodes:
            r,c = np.where(self.grid.tri.edges == i)
            for n in self.grid.tri.edges[r, c-1]:
                neighbors.add(n)
        return neighbors

    def taper_boundary(self, fld, depth=5):
        fld_taper = fld.copy()
        used_nodes = []
        nodes = self.get_boundary_nodes()
        fld_taper[nodes] = 0
        used_nodes.extend(nodes)
        for n in range(1, depth):
            nodes = self.get_neighbor_nodes(nodes)
            nodes.difference_update(used_nodes)
            nodes = list(set(nodes))
            fld_taper[nodes] *= n / depth
            used_nodes.extend(nodes)
        return fld_taper

    def displace(self, u, v, **kwargs):
        """
        Nextsim has a Lagrangian mesh, so it's possible to displace the mesh coordinates directly
        Inputs:
        - u, v: displacement vectors defined on self.grid.x,y
        """
        ##read grid, refresh self.grid.x, y
        self.read_grid(**kwargs)

        ##make sure boundary is not moving
        u = self.taper_boundary(u)
        v = self.taper_boundary(v)

        ##apply the displacement vectors -u, -v
        self.grid.x += u
        self.grid.y += v

        ##write the updated mesh node xy to the restart file
        self.write_grid(**kwargs)

    def read_mask(self, **kwargs):
        pass

    def prepare_mask(self, dst_grid):
        self.grid.set_destination_grid(dst_grid)
        tmp = self.grid.convert(np.ones(self.grid.x.shape))
        return np.isnan(tmp)

    def read_var(self, **kwargs):
        """read variable from a model restart file"""
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        if name in self.native_variables:
            var = read_data(fname, rec['name'])
            ##nextsim restart file concatenates u,v component, so reshape if is_vector
            if rec['is_vector']:
                var = var.reshape((2, -1))

        elif name in self.diag_variables:
            var = self.get_diag_var(**kwargs)

        elif name in self.atmos_forcing_variables:
            time = kwargs['time']
            nt_in_file = int(np.round(time.hour / rec['dt']))
            if rec['is_vector']:
                u = nc_read_var(fname, rec['name'][0])[nt_in_file, ...]
                v = nc_read_var(fname, rec['name'][1])[nt_in_file, ...]
                var = np.array([u, v])
            else:
                var = nc_read_var(fname, rec['name'])[nt_in_file, ...]

        ##convert units if native unit is not the same as required by kwargs
        var = units_convert(rec['units'], kwargs['units'], var)
        return var

    def write_var(self, var, **kwargs):
        """write variable back to a model restart file"""
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        name = kwargs['name']
        rec = self.variables[name]

        ##convert units back if necessary
        var = units_convert(kwargs['units'], rec['units'], var)

        if name in self.native_variables:
            ##nextsim restart file concatenates u,v component, so flatten if is_vector
            if rec['is_vector']:
                var = var.flatten()
            ##check if original var is on mesh nodes or elements
            # var_orig = read_data(fname, rec['name']).flatten()
            # if var_orig.size != var.size:
            #     ##the grid.convert interpolate to nodes by default, if size mismatch, this means
            #     ##we need element values, take the average of the node values here
            #     var = np.nanmean(var[grid.tri.triangles], axis=1)

            ##output the var to restart file
            write_data(fname, rec['name'], var)

        elif name in self.atmos_forcing_variables:
            time = kwargs['time']
            nt_in_file = int(np.round(time.hour / rec['dt']))
            ny, nx = var.shape[-2:]
            if rec['is_vector']:
                for i in range(2):
                    data_attr={'standard_name':rec['name'][i], 'units':rec['units'], 'grid_mapping':'projection_stereo'}
                    nc_write_var(fname, {'time':None, 'y':ny, 'x':nx}, rec['name'][i], var[i, ...], recno={'time':nt_in_file}, attr=data_attr, comm=kwargs['comm'])
            else:
                data_attr={'standard_name':rec['name'], 'units':rec['units'], 'grid_mapping':'projection_stereo'}
                nc_write_var(fname, {'time':None, 'y':ny, 'x':nx}, rec['name'], var, recno={'time':nt_in_file}, attr=data_attr, comm=kwargs['comm'])

    def z_coords(self, **kwargs):
        ##for nextsim, just discard inputs and simply return zero as z_coords
        return np.zeros(self.grid.x.shape)

    def get_seaice_drift(self, **kwargs):
        dt1day = timedelta(days=1)
        t2 = kwargs['time']
        t1 = kwargs['time'] - dt1day*3  ##TODO: make the duration configurable
        dt = (t2 - t1) / dt1day

        if kwargs['member'] is not None:
            mstr = '{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''

        meshfile = os.path.join(kwargs['path'], '..', '..', t2s(t1), 'nextsim.v1', mstr, 'restart', f"mesh_{t1:%Y%m%dT%H%M%SZ}.bin")
        self.read_grid(meshfile=meshfile, **kwargs)
        grid1 = self.grid_bank[meshfile]
        meshfile = os.path.join(kwargs['path'], '..', '..', t2s(t2-dt1day), 'nextsim.v1', mstr, 'restart', f"mesh_{t2:%Y%m%dT%H%M%SZ}.bin")
        self.read_grid(meshfile=meshfile, **kwargs)
        grid2 = self.grid_bank[meshfile]
        ids_cmn_12, ids1i, ids2i = np.intersect1d(grid1.id, grid2.id, return_indices=True)
        x1n = grid1.x[ids1i] / 1000
        y1n = grid1.y[ids1i] / 1000
        x2n = grid2.x[ids2i] / 1000
        y2n = grid2.y[ids2i] / 1000
        u = (x2n - x1n) / dt
        v = (y2n - y1n) / dt
        self.grid = Grid(grid1.proj, grid2.x[ids2i], grid2.y[ids2i], regular=False)
        return x2n, y2n, u, v

    def get_diag_var(self, **kwargs):
        name = kwargs['name']
        if name == 'seaice_drift':
            _, _, u, v = self.get_seaice_drift(**kwargs)
            return np.array([u, v])

        if 'seaice_deform' in name:
            x, y, u, v = self.get_seaice_drift(**kwargs)
            e1, e2, e3, _, _, _ = get_deformation_nodes(x, y, u, v)
            if name == 'seaice_deform_div':
                return e1
            if name == 'seaice_deform_shear':
                return e2
            if name == 'seaice_deform_vort':
                return e3

        raise NotImplementedError('cannot get diagnostic variable {name}')

    def read_param(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        return getattr(self, kwargs['name'])

    def write_param(self, param, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        setattr(self, kwargs['name'], param)

    def preprocess(self, task_id=0, **kwargs):
        """Preprocess the dir, collect input files for model run"""
        ##put sequence of operation here to generate the initial condition files for nextsim
        kwargs = super().parse_kwargs(**kwargs)
        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h

        if kwargs['member'] is not None:
            mstr = '{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''
        run_dir = os.path.join(kwargs['path'], mstr)
        makedir(run_dir)

        ##prepare restart files
        restart_file = self.filename(**{**kwargs, 'path':kwargs['restart_dir']})
        shell_cmd = f"cd {run_dir}; "
        shell_cmd += f"mkdir -p {self.restart_input_path}; cd {self.restart_input_path}; "
        field_bin = restart_file
        field_dat = field_bin.replace('.bin', '.dat')
        mesh_bin = restart_file.replace('field', 'mesh')
        mesh_dat = mesh_bin.replace('.bin', '.dat')
        for file in [field_bin, field_dat, mesh_bin, mesh_dat]:
            shell_cmd += f"cp -fL {file} .; "
        run_command(shell_cmd)

        ##prepare other input data (bathymetry, forcing, etc.) for the model run
        shell_cmd = f"cd {run_dir}; "
        shell_cmd += f"rm -rf data; mkdir -p data; cd data; "
        ##bathymetry data
        shell_cmd += f"ln -fs {os.path.join(self.nextsim_data_dir, 'BATHYMETRY', '*')} .; "
        ##ocean forcing (for now it is just linked over)
        shell_cmd += f"ln -fs {os.path.join(self.nextsim_data_dir, self.ocean_forcing_path)} .; "
        ##atmos forcing (make a copy, later they will be perturbed)
        shell_cmd += f"mkdir -p {self.atmos_forcing_path}; cd {self.atmos_forcing_path}; "
        t = time
        while t <= next_time:
            shell_cmd += f"cp -fL {os.path.join(self.nextsim_data_dir, self.atmos_forcing_path, 'generic_ps_atm_'+t.strftime('%Y%m%d')+'.nc')} .; "
            t += 24 * dt1h  ##forcing files are stored daily
        run_command(shell_cmd)

    def postprocess(self, task_id=0, **kwargs):
        ##place holder for now
        ##for any post processing needed after assimilation, to fix any model state that is not consistent
        kwargs = super().parse_kwargs(**kwargs)
        time = kwargs['time']
        if kwargs['member'] is not None:
            mstr = '{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''
        run_dir = os.path.join(kwargs['path'], mstr)
        restart_file = self.filename(**kwargs)

        ##read seaice conc and thick, check value, fix values out of normal range, then write back to file
        sic = self.read_var(**{**kwargs, 'name':'seaice_conc', 'units':1})
        sit = self.read_var(**{**kwargs, 'name':'seaice_thick', 'units':'m'})
        damage = self.read_var(**{**kwargs, 'name':'seaice_damage', 'units':1})
        rr = self.read_var(**{**kwargs, 'name':'seaice_ridge_ratio', 'units':1})

        sic = np.maximum(np.minimum(sic, 0.9999), 0.0)
        sit = np.maximum(sit, 0.0)
        damage = np.maximum(np.minimum(damage, 0.9999), 0.0)
        rr = np.maximum(np.minimum(rr, 0.9999), 0.0)

        self.write_var(sic, **{**kwargs, 'name':'seaice_conc', 'units':1})
        self.write_var(sit, **{**kwargs, 'name':'seaice_thick', 'units':'m'})
        self.write_var(damage, **{**kwargs, 'name':'seaice_damage', 'units':1})
        self.write_var(rr, **{**kwargs, 'name':'seaice_ridge_ratio', 'units':1})

    def run(self, task_id=0, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        self.run_status = 'running'

        time = kwargs['time']
        forecast_period = kwargs['forecast_period']
        next_time = time + forecast_period * dt1h
        input_file = self.filename(**kwargs)
        output_file = self.filename(**{**kwargs, 'time':next_time})

        if kwargs['member'] is not None:
            mstr = '{:03d}'.format(kwargs['member']+1)
        else:
            mstr = ''
        run_dir = os.path.join(kwargs['path'], mstr)
        makedir(run_dir)

        ##check input files
        field_bin = input_file
        field_dat = field_bin.replace('.bin', '.dat')
        mesh_bin = input_file.replace('field', 'mesh')
        mesh_dat = mesh_bin.replace('.bin', '.dat')
        for file in [field_bin, field_dat, mesh_bin, mesh_dat]:
            if not os.path.exists(file):
                raise RuntimeError("input file is missing: "+file)

        ##build command to run the model
        model_exe = os.path.join(self.nextsim_dir, 'model', 'bin', 'nextsim.exec')
        log_file = os.path.join(run_dir, 'run.log')
        run_command("touch "+log_file)

        shell_cmd = f". {self.model_env}; "
        shell_cmd += f"cd {run_dir}; "
        shell_cmd += f"export NEXTSIM_DATA_DIR={os.path.join(run_dir,'data')}; "
        shell_cmd += f"JOB_EXECUTE {model_exe} --config-files=config/nextsim.cfg >& run.log"

        ##give it several tries, each time decreasing time step
        for dt_ratio in [1, 0.5]:
            ##check output, if success skip further tries
            with open(log_file, 'rt') as f:
                if 'Simulation done' in f.read():
                    break

            self.timestep *= dt_ratio

            ##this creates nextsim.cfg.in in run_dir/config
            ##somehow the new version nextsim doesnt like nextsim.cfg to appear in run_dir
            config_dir = os.path.join(run_dir, 'config')
            makedir(config_dir)
            namelist(self, time, forecast_period, config_dir)

       ##run the model and wait for results
            run_job(shell_cmd, job_name='nextsim.run', run_dir=run_dir,
                    nproc=self.nproc_per_run, offset=task_id*self.nproc_per_run,
                    walltime=self.walltime, **kwargs)

        ##checkout output files
        watch_files([output_file])




import os

def value_str(value):
    ##convert values to string in namelist
    if isinstance(value, bool):
        if value:
            vstr = 'true'
        else:
            vstr = 'false'
    elif isinstance(value, str):
        vstr = value
    else:
        vstr = f'{value}'
    return vstr

def namelist(m, time, forecast_period, run_dir='.'):
    """Generate namelist for nextsim v1 model
    Input:
    -m: Model class object with model configurations
    -time: start time (datetime obj)
    -forecast_period: period (hour) to run the forecast
    -run_dir: directory where the model runtime files are stored
    """
    ##start building the namelist content
    nmlstr = "[mesh]\n"
    nmlstr += "filename="+value_str(m.msh_filename)+"\n"
    nmlstr += "[setup]\n"
    nmlstr += "ice-type="+value_str(m.ice_type)+"\n"
    nmlstr += "ocean-type="+value_str(m.ocean_type)+"\n"
    nmlstr += "atmosphere-type="+value_str(m.atmosphere_type)+"\n"
    nmlstr += "bathymetry-type="+value_str(m.bathymetry_type)+"\n"
    nmlstr += "use_assimilation=false\n"
    nmlstr += "dynamics-type="+value_str(m.dynamics_type)+"\n"
    nmlstr += "atmospheric_forcing_input_path="+os.path.join('data',m.atmos_forcing_path)+"\n"
    nmlstr += "oceanic_forcing_input_path="+os.path.join('data',m.ocean_forcing_path)+"\n"
    nmlstr += "[simul]\n"
    nmlstr += "spinup_duration=0\n"
    nmlstr += "timestep="+value_str(int(m.timestep))+"\n"
    nmlstr += "time_init="+time.strftime('%Y-%m-%d %H:%M:%S')+"\n"
    nmlstr += "duration="+f"{forecast_period / 24}"+"\n"
    nmlstr += "[thermo]\n"
    nmlstr += "use_assim_flux=false\n"
    nmlstr += "assim_flux_exponent=4\n"
    nmlstr += "diffusivity_sss="+value_str(m.diffusivity_sss)+"\n"
    nmlstr += "diffusivity_sst="+value_str(m.diffusivity_sst)+"\n"
    nmlstr += "ocean_nudge_timeS_days="+value_str(m.ocean_nudge_timeS)+"\n"
    nmlstr += "ocean_nudge_timeT_days="+value_str(m.ocean_nudge_timeT)+"\n"
    nmlstr += "freezingpoint-type="+m.freezingpoint_type+"\n"
    nmlstr += "h_young_max="+value_str(m.h_young_max)+"\n"
    nmlstr += "alb_ice="+value_str(m.alb_ice)+"\n"
    nmlstr += "alb_sn="+value_str(m.alb_sn)+"\n"
    nmlstr += "albedoW="+value_str(m.albedoW)+"\n"
    nmlstr += "melt_type="+value_str(m.melt_type)+"\n"
    nmlstr += "PhiM="+value_str(m.PhiM)+"\n"
    nmlstr += "snow_cond="+value_str(m.snow_cond)+"\n"
    nmlstr += "[dynamics]\n"
    nmlstr += "time_relaxation_damage="+value_str(m.time_relaxation_damage)+"\n"
    nmlstr += "compression_factor="+value_str(m.compression_factor)+"\n"
    nmlstr += "C_lab="+value_str(m.cohesion)+"\n"
    nmlstr += "substeps="+value_str(m.substeps)+"\n"
    nmlstr += "use_temperature_dependent_healing="+value_str(m.use_temperature_dependent_healing)+"\n"
    nmlstr += "ECMWF_quad_drag_coef_air="+value_str(m.ECMWF_quad_drag_coef_air)+"\n"
    nmlstr += "quad_drag_coef_water="+value_str(m.quad_drag_coef_water)+"\n"
    nmlstr += "Lemieux_basal_k1="+value_str(m.Lemieux_basal_k1)+"\n"
    nmlstr += "compaction_param="+value_str(m.compaction_param)+"\n"
    nmlstr += "oceanic_turning_angle="+value_str(m.oceanic_turning_angle)+"\n"
    nmlstr += "[age]\n"
    nmlstr += "include_young_ice="+value_str(m.include_young_ice)+"\n"
    nmlstr += "[restart]\n"
    nmlstr += "start_from_restart="+value_str(m.start_from_restart)+"\n"
    nmlstr += "type="+value_str(m.restart_type)+"\n"
    nmlstr += "write_initial_restart="+value_str(m.write_initial_restart)+"\n"
    nmlstr += "write_interval_restart="+value_str(m.write_interval_restart)+"\n"
    nmlstr += "output_interval="+f"{m.restart_dt / 24}"+"\n"
    nmlstr += "check_restart="+value_str(m.check_restart)+"\n"
    nmlstr += "input_path="+value_str(m.restart_input_path)+"\n"
    nmlstr += "basename="+time.strftime('%Y%m%dT%H%M%SZ')+"\n"
    nmlstr += "[output]\n"
    nmlstr += "output_per_day="+f"{int(24 / m.restart_dt)}"+"\n"
    nmlstr += "datetime_in_filename="+value_str(m.datetime_in_filename)+"\n"
    nmlstr += "exporter_path="+value_str(m.exporter_path)+"\n"
    nmlstr += "export_before_regrid="+value_str(m.export_before_regrid)+"\n"
    nmlstr += "export_after_regrid="+value_str(m.export_after_regrid)+"\n"
    nmlstr += "[moorings]\n"
    nmlstr += "use_moorings="+value_str(m.use_moorings)+"\n"
    nmlstr += "grid_type="+value_str(m.mooring_grid_type)+"\n"
    nmlstr += "spacing="+value_str(m.mooring_spacing)+"\n"
    nmlstr += "output_timestep="+f"{m.restart_dt / 24}"+"\n"
    for varname in m.mooring_variables:
        nmlstr += "variables="+varname+"\n"
    nmlstr += "[drifters]\n"
    nmlstr += "use_osisaf_drifters="+value_str(m.use_osisaf_drifters)+"\n"
    nmlstr += "use_sidfex_drifters="+value_str(m.use_sidfex_drifters)+"\n"
    nmlstr += "sidfex_drifters_output_time_step="+f"{m.sidfex_output_dt / 24}"+"\n"
    nmlstr += "sidfex_ignore_restart="+value_str(m.sidfex_ignore_restart)+"\n"
    nmlstr += "[debugging]\n"
    nmlstr += "check_fields_fast="+value_str(m.check_fields_fast)+"\n"

    ##write the namelist to nextsim.cfg file
    with open(os.path.join(run_dir, 'nextsim.cfg'), 'wt') as f:
        f.write(nmlstr)




import os
from pyproj import Proj
import numpy as np
from NEDAS.grid import Grid
from NEDAS.utils.conversion import units_convert, s2t, dt1h
from NEDAS.datasets.ecmwf import atmos_utils
from NEDAS.datasets import Dataset

from NEDAS.datasets.ecmwf.era5 import ERA5Data
era5 = ERA5Data()

class EcmwfForecastData(Dataset):
    def __init__(self, config_file=None, parse_args=False, **kwargs):
        super().__init__(config_file, parse_args, **kwargs)
        self.variables = {
            'atmos_surf_velocity': {'name':("10 metre U wind component", "10 metre V wind component"), 'is_vector':True, 'units':'m/s'},
            'atmos_surf_temp': {'name':"2 metre temperature", 'is_vector':False, 'units':'K'},
            'atmos_surf_dewpoint':  {'name':"2 metre dewpoint temperature", 'is_vector':False, 'units':'K'},
            'atmos_surf_press': {'name':"Mean sea level pressure", 'is_vector':False, 'units':'Pa'},
            'atmos_cloud_cover':  {'name':"Total cloud cover", 'is_vector':False, 'units':1},
            'atmos_precip': {'name':"Total precipitation", 'is_vector':False, 'units':'m/s'},
            'atmos_down_longwave': {'name':"Surface long-wave (thermal) radiation downwards", 'is_vector':False, 'units':'W/m2'},
            'atmos_down_shortwave': {'name':"Surface short-wave (solar) radiation downwards", 'is_vector':False, 'units':'W/m2'},
            'atmos_surf_vapor_mix': {'getter':self.get_vapmix, 'is_vector':False, 'units':'kg/kg'},
            }
        self.files = {}
        self.lookup = {}
        self.grid = None
        if isinstance(self.time_start, str):
            self.time_start = s2t(self.time_start)

    ###format filename
    def filename(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        t = self.time_start
        file = os.path.join(kwargs['path'], f"{t:%Y-%m}", f"fc_{t:%Y-%m-%d}.grb")
        assert os.path.exists(file), f"file {file} does not exist"
        return file

    def open_file(self, fname):
        if fname not in self.files:
            try:
                import pygrib
            except ImportError:
                raise RuntimeError("pygrib package is required to open ecmwf forecast data files.")
            print(f"opening file {fname}")
            self.files[fname] = pygrib.open(fname)
            self.get_record_id_lookup(fname)

    def close_file(self, fname):
        self.files[fname].close()

    def read_grid(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        fname = self.filename(**kwargs)
        self.open_file(fname)
        lat, lon = self.files[fname].message(1).latlons()
        self.grid = Grid(Proj('+proj=longlat'), lon, lat, cyclic_dim='x', pole_dim='y', pole_index=(0,))

    def get_record_id_lookup(self, fname):
        self.lookup[fname] = {}
        grbs = self.files[fname]
        for i in range(grbs.messages):
            grb = grbs.message(i+1)
            variable_name = grb.name
            forecast_hours = int(grb.stepRange)
            start_date = grb.analDate
            member = grb.perturbationNumber
            key = (variable_name, start_date, forecast_hours, member)
            self.lookup[fname][key] = i+1

    def read_data_from_grb(self, fname, time, member, vname, units):
        forecast_hours = int((time - self.time_start) / dt1h)
        ##build search key
        key = (vname, self.time_start, forecast_hours, member)
        ##look up the message id
        rec_id = self.lookup[fname][key]
        ##read the message into var
        grb = self.files[fname].message(rec_id)
        var = grb.values
        return var

    def read_var(self, **kwargs):
        kwargs = super().parse_kwargs(**kwargs)
        name = kwargs['name']
        assert name is not None, 'please specify which variable (name=?) to get'
        time = kwargs['time']
        k = kwargs.get('k', 0)
        member = kwargs['member']
        assert member is not None, 'please specify which member (member=?) to get'
        rec = self.variables[name]

        if 'name' in rec:
            fname = self.filename(**kwargs)
            self.open_file(fname)

            if rec['is_vector']:
                var1 = self.read_data_from_grb(fname, time, member, rec['name'][0], rec['units'])
                var2 = self.read_data_from_grb(fname, time, member, rec['name'][1], rec['units'])
                var = np.array([var1, var2])
            else:
                var = self.read_data_from_grb(fname, time, member, rec['name'], rec['units'])

            ##some variables are accumulated over the dt_hours, convert them to fluxes
            if name in ('atmos_precip', 'atmos_down_shortwave', 'atmos_down_longwave'):
                if time == self.time_start:
                    ##first time step is zeros for these variables
                    ##to ensure continuity in time, get the variable snapshot from ERA5 instead
                    era5.read_grid(name=name, time=time)
                    tmp = era5.read_var(name=name, time=time)
                    ##convert to grid
                    era5.grid.set_destination_grid(self.grid)
                    var = era5.grid.convert(tmp)
                else:
                    prev_time = time - dt1h * self.dt_hours
                    var -= self.read_data_from_grb(fname, prev_time, member, rec['name'], rec['units'])
                    ##need to convert the fluxes to per second units
                    var /= 3600. * self.dt_hours
                var[np.where(var<0.)] = 0. ##ensure positive definite
        else:
            var = rec['getter'](**kwargs)

        ##convert units if necessary
        var = units_convert(rec['units'], kwargs['units'], var)
        return var

    def get_vapmix(self, **kwargs):
        dewpoint = self.read_var(**{**kwargs, 'name':'atmos_surf_dewpoint', 'units':'K'})
        press = self.read_var(**{**kwargs, 'name':'atmos_surf_press', 'units':'Pa'})
        e = atmos_utils.satvap(dewpoint)
        vapmix = atmos_utils.vapmix(e, press)
        return vapmix

    def read_obs(self):
        raise NotImplementedError
